/* ==========================================================================
 * dns/db/file.c - DNS library.
 * --------------------------------------------------------------------------
 * Copyright (c) 2007  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */
#include <stdlib.h>	/* malloc(3) */
#include <stdio.h>	/* FILE getc(3) fclose(3) */

#include <ctype.h>	/* isspace(3) */

#include <string.h>	/* memchr(3) strcasecmp(3) memcmp(3) */

#include <sys/tree.h>	/* SPLAY_HEAD SPLAY_ENTRY SPLAY_GENERATE */
#include <sys/socket.h>	/* AF_INET AF_INET6 */

#include <netinet/in.h>	/* struct in_addr struct in6_addr */
#include <arpa/inet.h>	/* inet_pton(3) */

#include "dns.h"
#include "file.h"


#ifndef DNS_DB_FP__MAXDN
#define DNS_DB_FP__MAXDN	255
#endif

#ifndef DNS_DB_FP__TTL
#define DNS_DB_FP__TTL		180
#endif


struct dns_db_fp__entry {
	SPLAY_ENTRY(dns_db_fp__entry) res;
	SPLAY_ENTRY(dns_db_fp__entry) ptr;

	enum dns_rr_type rtype;

	union {
		struct dns_m_a a;
		struct dns_m_aaaa aaaa;
	} rr;

	unsigned char dn[];
}; /* struct dns_db_fp_entry */


struct dns_db_file {
	struct dns_database db;
	FILE *fp;

	SPLAY_HEAD(dns_db_fp__res, dns_db_fp__entry) res;
	SPLAY_HEAD(dns_db_fp__ptr, dns_db_fp__entry) ptr;
}; /* struct dns_db_file */


static int dns_db_fp__res_cmp(struct dns_db_fp__entry *a, struct dns_db_fp__entry *b) {
	int cmp;

	if (0 != (cmp = strcasecmp(a->dn, b->dn)))
		return cmp;

	return a->rtype - b->rtype;
} /* dns_db_fp__res_cmp() */

SPLAY_PROTOTYPE(dns_db_fp__res, dns_db_fp__entry, res, dns_db_fp__res_cmp);
SPLAY_GENERATE(dns_db_fp__res, dns_db_fp__entry, res, dns_db_fp__res_cmp);


static int dns_db_fp__ptr_cmp(struct dns_db_fp__entry *a, struct dns_db_fp__entry *b) {
	int cmp;

	if (0 != (cmp = a->rtype - b->rtype))
		return cmp;

	switch (a->rtype) {
	case DNS_RR_T_A:
		if (a->rr.a.addr < b->rr.a.addr)
			return -1;
		else if (a->rr.a.addr > b->rr.a.addr)
			return 1;
		else
			return 0;
	case DNS_RR_T_AAAA:
		return memcmp(a->rr.aaaa.addr, b->rr.aaaa.addr, sizeof a->rr.aaaa.addr);
	default:
		return -1;
	}
} /* dns_db_fp__ptr_cmp() */

SPLAY_PROTOTYPE(dns_db_fp__ptr, dns_db_fp__entry, ptr, dns_db_fp__ptr_cmp);
SPLAY_GENERATE(dns_db_fp__ptr, dns_db_fp__entry, ptr, dns_db_fp__ptr_cmp);


static void dns_db_fp__flush(struct dns_db_file *db) {
	struct dns_db_fp__entry *ent, *nxt;

	for (ent = SPLAY_MIN(dns_db_fp__res, &db->res); ent != 0; ent = nxt) {
		nxt	= SPLAY_NEXT(dns_db_fp__res, &db->res, ent);

		SPLAY_REMOVE(dns_db_fp__res, &db->res, ent);
		SPLAY_REMOVE(dns_db_fp__ptr, &db->ptr, ent);

		free(ent);
	}

	for (ent = SPLAY_MIN(dns_db_fp__ptr, &db->ptr); ent != 0; ent = nxt) {
		nxt	= SPLAY_NEXT(dns_db_fp__ptr, &db->ptr, ent);

		SPLAY_REMOVE(dns_db_fp__ptr, &db->ptr, ent);

		free(ent);
	}

	return /* void */;
} /* dns_db_fp__flush() */


static enum dns_errno dns_db_fp__push_entry(struct dns_db_file *db, char *ip, size_t iplen, char *dn, size_t dnlen) {
	struct dns_db_fp__entry *ent	= 0;
	struct in_addr in;
#ifdef AF_INET6
	struct in6_addr in6;
#endif

	if (0 == (ent = malloc(offsetof(struct dns_db_fp__entry, dn) + dnlen + 1)))
		goto anyfail;

	memcpy(ent->dn, dn, dnlen);
	ent->dn[dnlen]	= '\0';

#ifdef AF_INET6
	if (memchr(ip, ':', iplen) && 1 == inet_pton(AF_INET6, ip, &in6)) {
		ent->rtype	= DNS_RR_T_AAAA;
		memcpy(ent->rr.aaaa.addr, in6.s6_addr, sizeof ent->rr.aaaa.addr);
	} else
#endif
	if (1 == inet_pton(AF_INET, ip, &in)) {
		ent->rtype	= DNS_RR_T_A;
		ent->rr.a.addr	= in.s_addr;
	} else
		goto anyfail;

fprintf(stderr, "(%s) %.*s %.*s\n", (ent->rtype == DNS_RR_T_A)? "A" : "AAAA", (int)iplen, ip, (int)dnlen, dn);

	if (0 != SPLAY_INSERT(dns_db_fp__res, &db->res, ent)
	||  0 != SPLAY_INSERT(dns_db_fp__ptr, &db->ptr, ent))
		goto anyfail;

	return 0;
anyfail:
	free(ent);

	return 0;
} /* dns_db_fp__push_entry() */


static enum dns_errno dns_db_fp__parse(struct dns_db_file *dbf) {
	int ch;
	struct {
		char base[DNS_DB_FP__MAXDN + 1];
		char *pos, *end;
	} ip, dn, *sb;

	ip.end	= &ip.base[sizeof ip.base];
	dn.end	= &ip.base[sizeof dn.base];

nextip:
	ip.pos	= ip.base;
	sb	= &ip;
nextdn:
	dn.pos	= dn.base;
nextch:
	switch ((ch = getc(dbf->fp))) {
	case '#':
		do {
			ch = getc(dbf->fp);
		} while (ch != EOF && ch != '\n');

		/* FALL THROUGH */
	case EOF:
		/* FALL THROUGH */
	case '\n':
		break;
	default:
		if (isspace(ch)) {
			if (sb->base == sb->pos)
				goto nextch;

			break;
		}

		if (sb->pos < sb->end)
			*sb->pos++	= ch;

		goto nextch;
	} /* switch (getc()) */

	if (sb->pos != sb->base) {
		if (sb == &ip) {
			sb	= &dn;

			goto nextch;
		}

		if (ip.pos < ip.end - 1 && dn.pos < dn.end - 1) {
			*ip.pos	= '\0';
			*dn.pos	= '\0';

			(void)dns_db_fp__push_entry(dbf, ip.base, ip.pos - ip.base, dn.base, dn.pos - dn.base);
		}
	}

	if (ch == '\n')
		goto nextip;
	else if (ch != EOF)
		goto nextdn;

	return 0;
} /* dns_db_fp__parse() */


static void dns_db_fp__query(struct dns_db_file *db, struct dns_message *m0, void (*cb)(struct dns_database *, struct dns_message *, enum dns_errno, void *), void *arg, struct timeval timeout) {
	union {
		struct dns_db_fp__entry ent;
		unsigned char pad[offsetof(struct dns_db_fp__entry, dn) + DNS_DB_FP__MAXDN + 1];
	} k;
	struct dns_db_fp__entry *ent;
	struct dns_m_rr *rr;
	size_t len;
	struct dns_message *m	= 0;
	unsigned char dn[DNS_DB_FP__MAXDN];
	unsigned char rinfo[10];
	unsigned char rdata[sizeof ent->rr];
	unsigned short count	= 0;
	unsigned char *ancount;

	DNS_M_RR_FOREACH(rr, m, &m0->question, DNS_RR_CLASS(DNS_RR_C_IN), DNS_RR_TYPES(DNS_RR_T_A, DNS_RR_T_AAAA, DNS_RR_T_PTR)) {
		switch (k.ent.rtype = rr->type) {
		case DNS_RR_T_A:
		case DNS_RR_T_AAAA:
			len	= dns_m_dn_expand(m, k.ent.dn, DNS_DB_FP__MAXDN + 1, rr->name.base, DNS_M_DN_NOROOT);

			if (len == -1 || len > DNS_DB_FP__MAXDN)
				continue;

			if (0 != (ent = SPLAY_FIND(dns_db_fp__res, &db->res, &k.ent)))
				break;

#if 0
			if (m == 0) {
				if (0 == (m = dns_m_open(m0, 0)))
					goto anyfail;

				if (-1 == (len = dns_m_dn_comp(m, dn, sizeof dn, k.ent.dn, len, 0)))
					goto anyfail;

			}

			if (-1 == (len = dns_m_dn_comp(m, dn, sizeof dn, k.ent.dn, len, 0)))
				goto anyfail;

			if (0 != dns_m_parse(&m, dn, len, 0))
				goto anyfail;

			if (rr->type == DNS_RR_T_AAAA) {
				len	= sizeof ent->rr.aaaa.addr;
				memcpy(rdata, ent->rr.aaaa.addr, len);
			} else {
				unsigned int addr	= ntohl(ent->rr.a.addr);

				len		= 4;
				rdata[0]	= 0xff & (addr >> 24);
				rdata[1]	= 0xff & (addr >> 16);
				rdata[2]	= 0xff & (addr >> 8);
				rdata[3]	= 0xff & (addr >> 0);
			}

			rinfo[0]	= 0xff & (rr->type >> 8);
			rinfo[1]	= 0xff & (rr->type >> 0);
			rinfo[2]	= 0xff & (rr->class >> 8);
			rinfo[3]	= 0xff & (rr->class >> 0);
			rinfo[4]	= 0xff & (DNS_DB_FP__TTL >> 24);
			rinfo[5]	= 0xff & (DNS_DB_FP__TTL >> 16);
			rinfo[6]	= 0xff & (DNS_DB_FP__TTL >> 8);
			rinfo[7]	= 0xff & (DNS_DB_FP__TTL >> 0);
			rinfo[8]	= 0xff & (len >> 8);
			rinfo[9]	= 0xff & (len >> 0);

			if (0 != dns_m_parse(&m, rinfo, sizeof rinfo, 0)
			||  0 != dns_m_parse(&m, rdata, len, 0))
				goto anyfail;
#endif
			count++;

			continue;
		default:
			continue;
		}
	}

	if (0 == (ancount = dns_m__peek(m, m->data.base, DNS_M_OFF_ANCOUNT, 2)))
		goto anyfail;

	ancount[0]	= 0xff & (count >> 8);
	ancount[1]	= 0xff & (count >> 0);

	return /* void */;
anyfail:
	dns_m_close(m);

	return /* void */;
} /* dns_db_fp__query() */


static void dns_db_fp__close(struct dns_db_file *db) {
	if (db == 0)
		return /* void */;

	fclose(db->fp);

	dns_db_fp__flush(db);

	memset(db, 0, sizeof *db);

	free(db);

	return /* void */;
} /* dns_db_fp__close() */


struct dns_database *dns_db_file_open(FILE *fp, const struct dns_options *opts) {
	static const struct dns_db_file dbf_initializer	= {
		.db	= {
			.close	= (void *)&dns_db_fp__close,
			.query	= (void *)&dns_db_fp__query,
		},
	};
	struct dns_db_file *db;

	if (fp == 0 || 0 == (db = malloc(sizeof *db)))
		return dns_db__error(0, DNS_ESYSTEM), (void *)0;

	*db	= dbf_initializer;
	db->fp	= fp;

	SPLAY_INIT(&db->res);
	SPLAY_INIT(&db->ptr);

	dns_db_fp__parse(db);

	return &db->db;
} /* dns_db_file_open() */


