/* ==========================================================================
 * dns/message.c - DNS library.
 * --------------------------------------------------------------------------
 * Copyright (c) 2007  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */
#include <stddef.h>	/* size_t offsetof */
#include <stdint.h>	/* uintptr_t */
#include <stdbool.h>	/* bool */
#include <stdlib.h>	/* malloc(3) realloc(3) free(3) */

#include <string.h>	/* memset(3) memcpy(3) memcmp(3) */

#include "inline.h"

#if 1
#include <stdio.h>
#define MARK	fprintf(stderr, "@@%d\n", __LINE__)
#endif


#ifndef DNS_M_MINSIZE
#define DNS_M_MINSIZE	512
#endif

#ifndef DNS_M_MINRR
#define DNS_M_MINRR	16
#endif

#ifndef DNS_M_MINDICT
#define DNS_M_MINDICT	16	/* NB: Need at least 14 to fully compress "MX IN yahoo.com". */
#endif



static size_t dns_m__datasize(struct dns_message *m) {
	if (m->_.split)
		return roundup(m->length, DNS_M_MINSIZE);
	else
		return (unsigned char *)m->all.rr - m->data;
} /* dns_m__datasize() */


static size_t dns_m__rrtotal(struct dns_message *m) {
	if (m->_.split)
		return roundup(m->all.count, DNS_M_MINRR);
	else
		return (((char *)m + m->_.size) - (char *)m->all.rr) / sizeof *m->all.rr;
} /* dns_m__rrtotal() */


static void dns_m__rralign(struct dns_message *m) {
	m->qd.rr	= m->all.rr;
	m->an.rr	= &m->qd.rr[m->qd.count];
	m->ns.rr	= &m->an.rr[m->an.count];
	m->ar.rr	= &m->ns.rr[m->ns.count];
	m->dat.rr	= m->an.rr;

	return /* void */;
} /* dns_m__rralign() */


static struct dns_message *dns_m_grow(struct dns_message *m, size_t datalen, size_t nrr, enum dns_errno *err) {
	void *data;
	struct dns_record *rr;
	size_t sz;

	if (m == 0) {
		if (datalen > DNS_M_MAXSIZE || nrr > DNS_M_MAXRR)
			goto nomem;

		if (0 == (m = malloc(sz = sizeof *m + datalen + (sizeof *rr * (nrr + 1)))))
			goto nomem;

		memset(m, 0, sizeof *m);

		m->_.size	= sz;
		m->dict.limit	= DNS_M_MINDICT;
		m->data		= (void *)(m + 1);
		m->all.rr	= (void *)roundup((uintptr_t)&m->data[datalen], sizeof *rr);

		dns_m__rralign(m);

		return m;
	}

	if (dns_m__datasize(m) - m->length >= datalen && dns_m__rrtotal(m) - m->all.count >= nrr)
		return m;

	/* NB: Check integer overflows. */
	if (~m->length < datalen || ~m->all.count < nrr || m->length + datalen > DNS_M_MAXSIZE || m->all.count + nrr > DNS_M_MAXRR)
		goto nomem;

	if (m->_.split) {
		if (0 == (data = realloc(m->data, roundup(m->length + datalen, DNS_M_MINSIZE))))
			goto nomem;

		m->data		= data;

		if (0 == (rr = realloc(m->all.rr, sizeof *rr * roundup(m->all.count + nrr, DNS_M_MINRR))))
			goto nomem;

		m->all.rr	= rr;
	} else {
		data	= 0;
		rr	= 0;

		if (0 == (data = malloc(roundup(m->length + datalen, DNS_M_MINSIZE)))
		||  0 == (rr = malloc(sizeof *rr * roundup(m->all.count + nrr, DNS_M_MINRR)))) {
			free(data);
			free(rr);

			goto nomem;
		} else {
			m->data		= memcpy(data, m->data, m->length);
			m->all.rr	= memcpy(rr, m->all.rr, sizeof *rr * m->all.count);
		}

		m->_.split	= 1;
	}

	dns_m__rralign(m);

	return m;
nomem:
	*err	= DNS_ESYSTEM;

	return 0;
} /* dns_m_grow() */


struct dns_message *dns_m_parse(const void *data, size_t datalen, enum dns_errno *err) {
	unsigned short qdcount, ancount, nscount, arcount;
	unsigned int rrcount;
	struct dns_message *m;
	struct dns_record *rr;
	size_t p;

	if (datalen < DNS_M_HSIZE)
		return (*err = DNS_EINVALID), (void *)0;

	qdcount	= dns_unpack(DNS_M_H_QDCOUNT, data);
	ancount	= dns_unpack(DNS_M_H_ANCOUNT, data);
	nscount	= dns_unpack(DNS_M_H_NSCOUNT, data);
	arcount	= dns_unpack(DNS_M_H_ARCOUNT, data);

	rrcount	= qdcount + ancount + nscount + arcount;

	if (rrcount > DNS_M_MAXRR)
		return (*err = DNS_EINVALID), (void *)0;

	if (0 == (m = dns_m_grow(0, datalen, rrcount, err)))
		return 0;

	memcpy(m->data, data, datalen);
	m->length	= datalen;

	m->all.count	= rrcount;

	m->qd.count	= qdcount;
	m->an.count	= ancount;
	m->ns.count	= nscount;
	m->ar.count	= arcount;
	m->dat.count	= ancount + nscount + arcount;

	dns_m__rralign(m);

	for (p = DNS_M_HSIZE, rr = m->all.rr; p < m->length && m->all.found < m->all.count; m->all.found++, rr++) {
		if (0 == dns_rr_parse(rr, p, m->data, m->length, (m->all.found < m->qd.count)? DNS_RR_QR : 0, err))
			goto anyfail;

		p	+= dns_rr_length(rr);
	}

	m->qd.found	= (rr > m->qd.rr)? MIN(rr - m->qd.rr, m->qd.count) : 0;
	m->an.found	= (rr > m->an.rr)? MIN(rr - m->an.rr, m->an.count) : 0;
	m->ns.found	= (rr > m->ns.rr)? MIN(rr - m->ns.rr, m->ns.count) : 0;
	m->ar.found	= (rr > m->ar.rr)? MIN(rr - m->ar.rr, m->ar.count) : 0;
	m->dat.found	= (rr > m->dat.rr)? MIN(rr - m->dat.rr, m->dat.count) : 0;

	*err	= (m->all.found != m->all.count)? DNS_ETRUNCATED : 0;

	return m;
anyfail:
	dns_m_close(m);

	return 0;
} /* dns_m_parse() */


struct dns_message *dns_m_open(enum dns_errno *err) {
	struct dns_message *m;

	if (0 == (m = dns_m_grow(0, MAX(DNS_M_HSIZE, DNS_M_MINSIZE), DNS_M_MINRR, err)))
		return 0;

	memset(m->data, 0, DNS_M_HSIZE);
	m->length	= DNS_M_HSIZE;

	return m;
} /* dns_m_open() */


void dns_m_close(struct dns_message *m) {
	if (m == 0)
		return /* void */;

	if (m->_.split) {
		free(m->data);
		free(m->all.rr);
	}

	free(m->dict.name);
	free(m);

	return /* void */;
} /* dns_m_close() */


static void dns_m__rrsync(struct dns_message *m) {
	dns_pack(m->data, DNS_M_H_QDCOUNT, m->qd.count);
	dns_pack(m->data, DNS_M_H_ANCOUNT, m->an.count);
	dns_pack(m->data, DNS_M_H_NSCOUNT, m->ns.count);
	dns_pack(m->data, DNS_M_H_ARCOUNT, m->ar.count);

	return /* void */;
} /* dns_m__rrsync() */


struct dns_record *dns_m_push(struct dns_message *m, struct dns_section *s, const void *name_, size_t namelen, enum dns_rr_class class, enum dns_rr_type type, unsigned int ttl, const void *rdata, size_t rdlen, unsigned int flags, enum dns_errno *err) {
	struct { size_t datalen; unsigned short rrdelta; } save = { m->length, 0 };
	const unsigned char *name;
	unsigned char dn[DNS_DN_MAXSIZE];
	struct dns_record *rr;

	if (flags & DNS_M_NOCOMP) {
		name	= name_;
	} else {
		if (-1 == (namelen = dns_dn_comp(dn, sizeof dn, name_, namelen, dns_m_autodict(m), m->data, m->length, err)))
			return 0;

		name	= dn;
		rdlen	= 0;
	}

	if (~namelen < 10U || ~rdlen < namelen + 10U)
		return (*err = DNS_EINVALID), (void *)0;

	if (!dns_m_grow(m, namelen + 10 + rdlen, 1, err))
		return 0;

	if ((rr = &s->rr[s->count]) != &m->all.rr[m->all.count])
		return (*err = DNS_EINVALID), (void *)0;

	save.rrdelta++;

	m->all.count++;
	m->all.found++;
	s->count++;
	s->found++;

	dns_m__rralign(m);
	dns_m__rrsync(m);

	rr->dn.base	= m->length;
	rr->dn.length	= namelen;
	
	memcpy(&m->data[m->length], name, namelen);
	m->length	+= namelen;

	(void)dns_dict_appendbuf(&m->dict, rr->dn.base, m->data, m->length, m->data, m->length);

	rr->class	= class;
	rr->type	= type;

	dns_pack(&m->data[m->length], 0, 16, rr->type);
	dns_pack(&m->data[m->length], 16, 16, rr->class);

	m->length		+= 4;

	if ((rr->qr = (s == &m->qd))) {
		rr->ttl		= 0;
		rr->rd.base	= 0;
		rr->rd.length	= 0;

		return rr;
	}

	rr->ttl		= (0x7fffffff & ttl);

	dns_pack(&m->data[m->length], 0, 32, rr->ttl);

	m->length	+= 4;

	if (flags & DNS_M_NOCOMP) {
		dns_pack(&m->data[m->length], 0, 16, rdlen);

		m->length	+= 2;

		rr->rd.base	= m->length;
		rr->rd.length	= rdlen;

		memcpy(&m->data[m->length], rdata, rdlen);
		m->length	+= rdlen;
	} else {
		const struct dns_rr_parser *p;

		if (0 == (p = dns_rr_getparser(rr->class, rr->type, err)))
			goto rollback;

		if (!dns_m_grow(m, 2, 0, err))
			goto rollback;

		m->length	+= 2;

		do {
			if (!dns_m_grow(m, rdlen, 0, err))
				goto rollback;

			/*
			 * NB: Pass total available size as the "data"
			 * region so parsers can append compressed RDATA DNs
			 * to the compression dictionary. 
			 *
			 * dns_dict_appendbuf() uses this memory region
			 * logic to know whether to parse and append a
			 * compressed DN.
			 */
			if (-1 == (rdlen = p->comp(&m->data[m->length], dns_m__datasize(m) - m->length, rdata, dns_m_autodict(m), m->data, dns_m__datasize(m), err)))
				goto rollback;
		} while (rdlen > dns_m__datasize(m));

		dns_pack(&m->data[m->length - 2], 0, 16, rdlen);
		m->length	+= rdlen;
	}

	return &s->rr[s->count - 1];
rollback:
	m->all.count	-= save.rrdelta;
	m->all.found	-= save.rrdelta;
	s->count	-= save.rrdelta;
	s->found	-= save.rrdelta;

	dns_m__rralign(m);
	dns_m__rrsync(m);

	m->length	= save.datalen;

	/* FIXME: How do we undo compression dictionary additions? */
	free(m->dict.name);
	m->dict.name	= 0;
	m->dict.count	= 0;

	return 0;
} /* dns_m_push() */


struct dns_record *dns_m_next(struct dns_message *m, struct dns_section *s, const void *name_, size_t namelen, const struct dns_enumset *cset, const struct dns_enumset *tset, unsigned *state) {
	struct dns_record *rr	= 0;
	unsigned rri;

	rri	= *state;

	for (; rri < s->found && rr == 0; rri++) {
		/* FIXME: Match DN */

		if (!dns_enumset_has(cset, s->rr[rri].class) || !dns_enumset_has(tset, s->rr[rri].type))
			continue;

		rr	= &s->rr[rri];
	}

	*state	= rri;

	return rr;
} /* dns_m_next() */


struct dns_record *dns_rr_parse(struct dns_record *rr, unsigned short base, const void *data_, size_t datalen, unsigned int flags, enum dns_errno *err) {
	const unsigned char *data	= data_;
	size_t p			= base;
	struct dns_name dn;

	if (0 == dns_dn_parse(&dn, p, data, datalen, err))
		return 0;

	rr->dn.base	= dn.base;
	rr->dn.length	= dn.length;
		
	p	+= dn.length;

	if (p > datalen || datalen - p < 4)
		return (*err = DNS_ETRUNCATED), (void *)0;

	rr->type	= dns_unpack(0, 16, &data[p]);
	rr->class	= dns_unpack(16, 16, &data[p]);

	p	+= 4;

	if ((rr->qr = !!(flags & DNS_RR_QR))) {
		rr->ttl		= 0;
		rr->rd.base	= 0;
		rr->rd.length	= 0;

		return rr;
	}

	if (p > datalen || datalen - p < 6)
		return (*err = DNS_ETRUNCATED), (void *)0;

	rr->ttl		= dns_unpack(0, 32, &data[p]);
	rr->rd.length	= dns_unpack(32, 16, &data[p]);

	p	+= 6;

	rr->rd.base	= p;

	if (p > datalen || datalen - p < rr->rd.length)
		return (*err = DNS_ETRUNCATED), (void *)0;

	return rr;
} /* dns_rr_parse() */


size_t dns_dn_concat(void *dst_, size_t dstsiz, const struct dns_name *dn, unsigned char l, unsigned char n, const void *data_, size_t datalen, unsigned int flags, enum dns_errno *err) {
	unsigned char *dst		= dst_;
	size_t dstp			= 0;
	const unsigned char *data	= data_;
	unsigned int i;

	for (i = l; i < dn->count && n > 0; i++, n--) {
		if (dstp < dstsiz)
			memmove(&dst[dstp], &data[dn->label[i].base + 1], MIN(dstsiz - dstp, 0x3f & data[dn->label[i].base]));

		dstp	+= (0x3f & data[dn->label[i].base]);

		if (dstp < dstsiz)
			dst[dstp]	= '.';

		dstp++;
	}

	if ((flags & DNS_DN_NOROOT) && dstp > 0)
		dstp--;

	if (dstsiz > 0)
		dst[MIN(dstp, dstsiz - 1)]	= '\0';

	return dstp;
} /* dns_dn_concat() */


size_t dns_dn_expand(void *dst, size_t dstsiz, unsigned short datap, const void *data, size_t datalen, unsigned int flags, enum dns_errno *err) {
	struct dns_name dn;

	if (0 == dns_dn_parse(&dn, datap, data, datalen, err))
		return -1;

	return dns_dn_concat(dst, dstsiz, &dn, 0, -1, data, datalen, flags, err);
} /* dns_dn_expand() */


size_t dns_dn_comp(void *dst_, size_t dstlen, const void *name_, size_t namelen, struct dns_dict *dict, const void *data_, size_t datalen, enum dns_errno *err) {
	unsigned char *dst		= dst_;
	size_t p			= 0;
	const unsigned char *name	= name_;
	struct dns_dname dn;
	unsigned int i;
	unsigned short ptr;

	if (0 == dns_ddn_parse(&dn, name_, namelen, err))
		return -1;

	namelen	= dn.length;

	for (i = 0; i < dn.count; i++) {
		if (0 != (ptr = dns_dict_dsearch(dict, &dn, i, name_, namelen, data_, datalen))) {
			if (p < dstlen)
				dst[p]	= (0xc0 | (0x3f & (ptr >> 8)));

			p++;

			if (p < dstlen)
				dst[p]	= (0xff & (ptr >> 0));

			return ++p;
		}

		/*
		 * NB: Handle delicately so dst and src can overlap.
		 *
		 * 	(1) Conceptually, we're swapping the label separator
		 * 	    '.' for a prefix run length.
		 *
		 *	(2) Re-order the label copy so we don't lose
		 * 	    the first character in the label when we
		 * 	    write the prefix.
		 *
		 * 	(3) Use memmove(3), obviously.
		 *
		 *	(4) !!! This compression function will *expand*
		 * 	    !!! the input name if it lacks a trailing root
		 * 	    !!! separator (and it can't use pointers for any
		 * 	    !!! multi-character label). So, counter
		 * 	    !!! intuitively, the dst buffer should always be
		 * 	    !!! larger than the input name for good measure
		 * 	    !!! (up to DNS_DN_MAXSIZE, of course, and
		 * 	    !!! other extrinsic limitations).
		 */
		if (p < dstlen) {
			memmove(&dst[p + 1], &name[dn.label[i].base], MIN(dstlen - (p + 1), dn.label[i].length));
			dst[p]	= (0x3f & dn.label[i].length);
		}

		p	+= 1 + (0x3f & dn.label[i].length);
	}

	if (p < dstlen)
		dst[p]	= '\0';

	return ++p;
} /* dns_dn_comp() */


struct dns_name *dns_dn_parse(struct dns_name *dn, unsigned short base, const void *data_, size_t datalen, enum dns_errno *err) {
	const unsigned char *data	= data_;
	size_t p			= base;

	dn->base	= base;
	dn->length	= 0;
	dn->count	= 0;

	while (p < datalen) {
		switch (0x03 & (data[p] >> 6)) {
		case 0x00:	/* FOLLOWS */
			dn->label[dn->count].base	= p;

			if (0 == (0x3f & data[p])) {
				if (dn->length == 0)
					dn->length	= (p + 1) - base;

/* success ==> */		return dn;
			} else if (datalen - p <= (0x3f & data[p])) {
				goto truncat;
			} else
				p	+= (0x3f & data[p]) + 1;

			if (++dn->count >= arraylen(dn->label))
				goto invalid;

			continue;
		case 0x01:	/* RESERVED */
			goto invalid;
		case 0x02:	/* RESERVED */
			goto invalid;
		case 0x03:	/* POINTER */
			if (datalen - p < 2)
				goto truncat;

			if (dn->length == 0)
				dn->length	= (p + 2) - base;

			p	= ((0x3f & data[p + 0]) << 8)
				| ((0xff & data[p + 1]) << 0);

			continue;
		} /* switch (data[p]) */
	} /* while (p < datalen) */

truncat:
	return (*err = DNS_ETRUNCATED), (void *)0;
invalid:
	return (*err = DNS_EINVALID), (void *)0;
} /* dns_dn_parse() */


static inline bool dns_dict_labeleq(unsigned short a, unsigned short b, const unsigned char *data) {
	if ((0x3f & data[a]) != (0x3f & data[b]))
		return 0;

	return 0 == memcmp(&data[a + 1], &data[b + 1], (0x3f & data[a]));
} /* dns_dict_labeleq() */


enum dns_errno dns_dict_append(struct dns_dict *d, struct dns_name *dn, const void *data, size_t datalen) {
	void *tmp;
	unsigned short i, j, n;

	if (dn->count == 0 || dn->label[dn->count - 1].base > DNS_DN_MAXPTR)
		return 0;

	n	= 0;

	while (n < d->count) {
#if 0
unsigned char n1[256], n2[256];
dns_dn_concat(n1, sizeof n1, &d->name[n], 0, -1, data, datalen, 0, DNS_NOERRNO);
dns_dn_concat(n2, sizeof n2, dn, 0, -1, data, datalen, 0, DNS_NOERRNO);
fprintf(stderr, "[%s] <=> [%s]\n", n1, n2);
#endif
		if (d->name[n].count > dn->count)
			goto nextdn;

		for (i = 0, j = dn->count - d->name[n].count; i < d->name[n].count; i++, j++) {
			if (!dns_dict_labeleq(d->name[n].label[i].base, dn->label[j].base, data))
				goto nextdn;
		}

		d->name[n]	= *dn;

		return 0;
nextdn:
		n++;
	}

append:
	if (d->count >= d->limit || ~d->count < 1U)
		return 0;

	if (0 == (tmp = realloc(d->name,  sizeof *dn * (d->count + 1))))
		return DNS_ESYSTEM;

	d->name			= tmp;
	d->name[d->count++]	= *dn;

	return 0;
} /* dns_dict_append() */


enum dns_errno dns_dict_appendbuf(struct dns_dict *d, unsigned short dstp, const void *dst_, size_t dstsiz, const void *data_, size_t datalen) {
	const unsigned char *dst	= dst_;
	const unsigned char *data	= data_;
	struct dns_name dn;
	enum dns_errno err;

	if (dst < data || dstp > dstsiz || dst + dstsiz > data + datalen)
		return 0;

	if (!dns_dn_parse(&dn, &dst[dstp] - data, data, datalen, &err))
		return err;

	return dns_dict_append(d, &dn, data, datalen);
} /* dns_dict_appendbuf() */


unsigned short dns_dict_search(struct dns_dict *d, struct dns_name *dn, unsigned short l, const void *data, size_t datalen) {
	unsigned short i, j, n;

	n	= 0;

	while (n < d->count) {
#if 0
unsigned char n1[256], n2[256];
dns_dn_concat(n1, sizeof n1, &d->name[n], 0, -1, data, datalen, 0, DNS_NOERRNO);
dns_dn_concat(n2, sizeof n2, dn, l, -1, data, datalen, 0, DNS_NOERRNO);
fprintf(stderr, "[%s] <=> [%s]\n", n1, n2);
#endif
		if (d->name[n].count < (dn->count - l))
			goto nextdn;

		for (i = d->name[n].count - (dn->count - l), j = l; i < d->name[n].count; i++, j++) {
			if (!dns_dict_labeleq(d->name[n].label[i].base, dn->label[j].base, data))
				goto nextdn;
		}

		return d->name[n].label[d->name[n].count - (dn->count - l)].base;
nextdn:
		n++;
	}

	return 0;	
} /* dns_dict_search() */


static inline bool dns_dict_dlabeleq(unsigned short a, const void *b, size_t blen, const unsigned char *data) {
	if ((0x3f & data[a]) != blen)
		return 0;

	return 0 == memcmp(&data[a + 1], b, blen);
} /* dns_dict_dlabeleq() */


unsigned short dns_dict_dsearch(struct dns_dict *d, struct dns_dname *dn, unsigned short l, const void *name_, size_t namelen, const void *data, size_t datalen) {
	const unsigned char *name	= name_;
	unsigned int i, j, n;

	n	= 0;

	while (n < d->count) {
#if 0
unsigned char n1[256], n2[256];
dns_dn_concat(n1, sizeof n1, &d->name[n], 0, -1, data, datalen, 0, DNS_NOERRNO);
dns_ddn_concat(n2, sizeof n2, dn, l, -1, name, namelen, 0, DNS_NOERRNO);
fprintf(stderr, "[%s] <=> [%s]\n", n1, n2);
#endif
		if (d->name[n].count < (dn->count - l))
			goto nextdn;

		for (i = d->name[n].count - (dn->count - l), j = l; i < d->name[n].count; i++, j++) {
			if (!dns_dict_dlabeleq(d->name[n].label[i].base, &name[dn->label[j].base], dn->label[j].length, data))
				goto nextdn;
		}

		return d->name[n].label[d->name[n].count - (dn->count - l)].base;
nextdn:
		n++;
	}

	return 0;
} /* dns_dict_dsearch() */


struct dns_dname *dns_ddn_parse(struct dns_dname *dn, const void *name_, size_t namelen, enum dns_errno *err) {
	const unsigned char *name	= name_;
	size_t p			= 0;

	dn->length	= 0;
	dn->count	= 0;

	dn->label[0].base	= p;
	dn->label[0].length	= 0;

	for (; p < namelen && name[p] != '\0'; p++) {
		if (name[p] == '.') {
			if (dn->label[dn->count].length > 0) {
				if (++dn->count >= arraylen(dn->label))
					goto invalid;

				dn->label[dn->count].base	= p + 1;
				dn->label[dn->count].length	= 0;
			}
		} else {
			if (++dn->label[dn->count].length == 0)	/* OVERFLOW */
				goto invalid;
		}
	}

	if (dn->label[dn->count].length > 0)
		dn->count++;

	dn->length	= p;

	return dn;
invalid:
	return (*err = DNS_EINVALID), (void *)0;
} /* dns_ddn_parse() */


size_t dns_ddn_concat(void *dst_, size_t dstsiz, const struct dns_dname *dn, unsigned char l, unsigned char n, const void *data_, size_t datalen, unsigned int flags, enum dns_errno *err) {
	unsigned char *dst		= dst_;
	size_t dstp			= 0;
	const unsigned char *data	= data_;
	unsigned int i;

	for (i = l; i < dn->count && n > 0; i++, n--) {
		if (dstp < dstsiz)
			memmove(&dst[dstp], &data[dn->label[i].base], MIN(dstsiz - dstp, dn->label[i].length));

		dstp	+= dn->label[i].length;

		if (dstp < dstsiz)
			dst[dstp]	= '.';

		dstp++;
	}

	if ((flags & DNS_DN_NOROOT) && dstp > 0)
		dstp--;

	if (dstsiz > 0)
		dst[MIN(dstp, dstsiz - 1)]	= '\0';

	return dstp;
} /* dns_ddn_concat() */


