/* ==========================================================================
 * dns/message.c - DNS library.
 * --------------------------------------------------------------------------
 * Copyright (c) 2007  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */
#include <stddef.h>	/* size_t */


#define DNS_M_MAXSIZE		65535	/* 16-bit size limitation for TCP VC. */
#define DNS_M_MAXRR		((DNS_M_MAXSIZE - DNS_M_HSIZE) / DNS_RR_MINSIZE)

#define DNS_RR_MINSIZE		11	/* Minimum RR length. */

#define DNS_DN_MAXSIZE		255	/* Maximum size including NUL for wire format and unencoded ascii (255 per RFC 1035; 253 is longest ascii name which can fit in wire format). */
#define DNS_DN_MAXLABEL		127	/* Maximum # of labels/DN (254/2), not including root zone. */
#define DNS_DN_MAXPTR		16383	/* Maximum label pointer for compressed DN (1<<14 - 1). */

#define DNS_LB_MAXSIZE		63	/* Maximum length of label (not including NUL) */

#define DNS_M_HSIZE		12	/* Fixed size of DNS header. */

#define DNS_M_H_QID		0, 16	/* Header bit offset, length for */
#define DNS_M_H_QR		16, 1	/* use with dns_pack/dns_unpack. */
#define DNS_M_H_OPCODE		17, 4
#define DNS_M_H_AA		21, 1
#define DNS_M_H_TC		22, 1
#define DNS_M_H_RD		23, 1
#define DNS_M_H_RA		24, 1
#define DNS_M_H_Z		25, 3
#define DNS_M_H_RCODE		28, 4
#define DNS_M_H_QDCOUNT		32, 16
#define DNS_M_H_ANCOUNT		48, 16
#define DNS_M_H_NSCOUNT		64, 16
#define DNS_M_H_ARCOUNT		80, 16


enum dns_m_flags {
	DNS_M_NOCOMP	= (1U << 0),	/* Passed DN & RDATA already in wire format. */
}; /* enum dns_m_flags */


enum dns_rr_flags {
	DNS_RR_QR	= (1U << 8),	/* Parse QUERY record (no TTL, RDATA). */
}; /* enum dns_rr_flags */


enum dns_dn_flags {
	DNS_DN_NOROOT	= (1U << 16),	/* No root zone in ascii expansion. */
}; /* enum dns_dn_flags */


struct dns_name;
struct dns_dname;
struct dns_record;

struct dns_dict {
	unsigned short limit;
	unsigned short count;

	struct dns_name *name;
}; /* struct dns_dict */


struct dns_message {
	unsigned short qid;

	unsigned short qr:1;
	unsigned short opcode:4;
	unsigned short aa:1;
	unsigned short tc:1;
	unsigned short rd:1;
	unsigned short ra:1;
	unsigned short z:3;
	unsigned short rcode:4;

	struct dns_section {
		unsigned short base;
		unsigned short length;

		unsigned short count;
		unsigned short found;

		struct dns_record *rr;
	} all, qd, an, ns, ar, dat;

	struct dns_dict dict;

	unsigned char *data;
	unsigned short length;

	struct {
		size_t size;
		bool split;
	} _;
}; /* struct dns_message */


struct dns_message *dns_m_parse(const void *, size_t, enum dns_errno *);

struct dns_message *dns_m_open(enum dns_errno *);

void dns_m_close(struct dns_message *);

struct dns_record *dns_m_push(struct dns_message *, struct dns_section *, const void *, size_t, enum dns_rr_class, enum dns_rr_type, unsigned int ttl, const void *, size_t, unsigned int, enum dns_errno *);

#define DNS_DN_ANY	0, 0

#define DNS_M_FOREACH(...)	DNS_M_FOREACH_(__VA_ARGS__)
#define DNS_M_FOREACH_(rr, m, s, d, n, c, t) \
	for (unsigned dns_m_next__i = 0; 0 != ((rr) = dns_m_next((m), (s), (d), (n), (c), (t), &dns_m_next__i));)

struct dns_record *dns_m_next(struct dns_message *, struct dns_section *, const void *, size_t, const struct dns_enumset *, const struct dns_enumset *, unsigned *);

#define dns_m_find(...)	dns_m_next(__VA_ARGS__, &(unsigned int){ 0 })

#define dns_m_autodict(m)	(((m)->dict.count)? &(m)->dict : dns_dict_auto(DNS_DICT_AUTO(), dns_dn_parse(DNS_NAME_AUTO(), DNS_M_HSIZE, m->data, m->length, DNS_NOERRNO)))


struct dns_record {
	struct {
		unsigned short base;
		unsigned short length;
	} dn;

	enum dns_rr_type type;
	enum dns_rr_class class;
	unsigned int qr:1;
	unsigned int ttl:31;

	struct {
		unsigned short base;
		unsigned short length;
	} rd;
}; /* struct dns_record */

static inline size_t dns_rr_length(const struct dns_record *rr) {
	return (rr->qr)? rr->dn.length + 4 : rr->dn.length + 10 + rr->rd.length;
} /* dns_rr_length() */

struct dns_record *dns_rr_parse(struct dns_record *, unsigned short, const void *, size_t, unsigned int, enum dns_errno *);


#define DNS_NAME_AUTO()	(&(struct dns_name){ 0 })

struct dns_name {
	unsigned short base;
	unsigned short length;

	unsigned short count;

	struct {
		unsigned short base;
	} label[DNS_DN_MAXLABEL];
}; /* struct dns_name */

struct dns_name *dns_dn_parse(struct dns_name *, unsigned short, const void *, size_t, enum dns_errno *);
size_t dns_dn_concat(void *, size_t, const struct dns_name *, unsigned char, unsigned char, const void *, size_t, unsigned int, enum dns_errno *);
size_t dns_dn_expand(void *, size_t, unsigned short, const void *, size_t, unsigned int, enum dns_errno *);
size_t dns_dn_comp(void *, size_t, const void *, size_t, struct dns_dict *, const void *, size_t, enum dns_errno *);


#define DNS_DICT_AUTO()	(&(struct dns_dict){ 0 })

static inline struct dns_dict *dns_dict_auto(struct dns_dict *d, struct dns_name *dn) {
	d->limit	= 0;
	d->count	= !!dn;
	d->name		= dn;

	return d;
}; /* dns_dict_auto() */

enum dns_errno dns_dict_append(struct dns_dict *, struct dns_name *, const void *, size_t);
enum dns_errno dns_dict_appendbuf(struct dns_dict *, unsigned short, const void *, size_t, const void *, size_t);
unsigned short dns_dict_search(struct dns_dict *, struct dns_name *, unsigned short, const void *, size_t);
unsigned short dns_dict_dsearch(struct dns_dict *, struct dns_dname *, unsigned short, const void *, size_t, const void *, size_t);


struct dns_dname {
	unsigned short length;
	unsigned short count;

	struct {
		unsigned short base;
		unsigned short length;
	} label[DNS_DN_MAXLABEL];
}; /* struct dns_dname */

struct dns_dname *dns_ddn_parse(struct dns_dname *, const void *, size_t, enum dns_errno *);
size_t dns_ddn_concat(void *, size_t, const struct dns_dname *, unsigned char, unsigned char, const void *, size_t, unsigned int, enum dns_errno *);

