/* ==========================================================================
 * dns/rr.c - DNS library.
 * --------------------------------------------------------------------------
 * Copyright (c) 2007  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */
#include <stdlib.h>	/* malloc(3) */

#include <sys/tree.h>	/* RB */


struct dns_rr_p {
	enum dns_rr_class class;
	enum dns_rr_type type;

	struct dns_rr_parser parser;

	RB_ENTRY(dns_rr_p) rbe;
}; /* struct dns_rr_p */


static const struct dns_rr_parser dns_rr_builtin[] = {
	[DNS_RR_T_A]	= { &dns_rr_a_parse, &dns_rr_a_comp },
	[DNS_RR_T_AAAA]	= { &dns_rr_aaaa_parse, &dns_rr_aaaa_comp },
	[DNS_RR_T_MX]	= { &dns_rr_mx_parse, &dns_rr_mx_comp },
	[DNS_RR_T_NS]	= { &dns_rr_ns_parse, &dns_rr_ns_comp },
	[DNS_RR_T_TXT]	= { &dns_rr_txt_parse, &dns_rr_txt_comp },
}; /* dns_rr_builtin[] */


RB_HEAD(dns_rr_parsers, dns_rr_p);

static int dns_rr_pcmp(struct dns_rr_p *a, struct dns_rr_p *b) {
	int cmp;

	if (0 != (cmp = a->class - b->class) || 0 != (cmp = a->type - b->type))
		return cmp;

	return 0;
} /* dns_rr_pcmp() */

RB_PROTOTYPE(dns_rr_parsers, dns_rr_p, rbe, dns_rr_pcmp)
RB_GENERATE(dns_rr_parsers, dns_rr_p, rbe, dns_rr_pcmp)

static struct dns_rr_parsers dns_rr_parsers = RB_INITIALIZER(&dns_rr_parsers);


const struct dns_rr_parser *dns_rr_getparser(enum dns_rr_class c, enum dns_rr_type t, enum dns_errno *err) {
	unsigned i;
	struct dns_rr_p k, *p;

	if (c == DNS_RR_C_IN && t >= 0 && t < DNS__ARRAYLEN(dns_rr_builtin)) {
		if (dns_rr_builtin[t].parse != 0)
			return &dns_rr_builtin[t];
	}

	k.class	= c;
	k.type	= t;
	p	= RB_FIND(dns_rr_parsers, &dns_rr_parsers, &k);

	return (p)? &p->parser : ((*err = DNS_EINVALID), (void *)0);
} /* dns_rr_getparser() */


const struct dns_rr_parser *dns_rr_addparser(enum dns_rr_class class, enum dns_rr_type type, dns_rr_parse_fn pfn, dns_rr_comp_fn cfn, enum dns_errno *err) {
	struct dns_rr_p *p, *q;

	if (0 == (p = malloc(sizeof *p)))
		return (*err = DNS_ESYSTEM), (void *)0;

	p->class	= class;
	p->type		= type;
	p->parser.parse	= pfn;
	p->parser.comp	= cfn;

	if (0 != (q = RB_INSERT(dns_rr_parsers, &dns_rr_parsers, p))) {
		q->parser	= p->parser;

		free(p);

		return &q->parser;
	} else
		return &p->parser;
} /* dns_rr_addparser() */


size_t dns_rr_expand(void *name, size_t namelen, void *rr, size_t rrlen, unsigned short datap, const void *data_, size_t datalen, unsigned int flags, enum dns_errno *err) {
	const unsigned char *data	= data_;
	size_t namep, rdlen;
	struct dns_name dn;
	const struct dns_rr_parser *p;
	enum dns_rr_class class;
	enum dns_rr_type type;
	unsigned i;

	if (0 == dns_dn_parse(&dn, datap, data_, datalen, err))
		return -1;

	if (-1 == (namelen = dns_dn_concat(name, namelen, &dn, 0, -1, data, datalen, flags, err)))
		return -1;

	datap	+= dn.length;

	if (datap > datalen || datalen - datap < 4)
		goto truncat;

	type	= dns_unpack(0, 16, &data[datap]);
	class	= dns_unpack(16, 16, &data[datap]);

	if (flags & DNS_RR_QR)
		return namelen;

	datap	+= 4 + 4;	/* Skip TTL, too. */

	if (datap > datalen || datalen - datap < 2)
		goto truncat;

	rdlen	= dns_unpack(0, 16, &data[datap]);
	datap	+= 2;

	if (datap > datalen || datalen - datap < rdlen)
		goto truncat;

	if (0 == (p = dns_rr_getparser(class, type, err)))
		goto unknown;

	if (!p->parse(rr, rrlen, datap, rdlen, data, datalen, flags, err))
		return -1;

	return namelen;
unknown:
truncat:
	return -1;
} /* dns_rr_expand() */


size_t dns_rr_comp(void *dst_, size_t dstlen, const void *name_, size_t namelen, enum dns_rr_class class, enum dns_rr_type type, unsigned int ttl, const void *rr, struct dns_dict *d, const void *data, size_t datalen, enum dns_errno *err) {
	return -1;
} /* dns_rr_comp() */

