/* ==========================================================================
 * dns/rr/txt.c - DNS library.
 * --------------------------------------------------------------------------
 * Copyright (c) 2007  William Ahern
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 * USE OR OTHER DEALINGS IN THE SOFTWARE.
 * ==========================================================================
 */
#include <stdarg.h>	/* va_list va_start va_end */
#include <stdio.h>	/* vsnprintf(3) */

#include <string.h>	/* memcpy(3) memmove(3) */


void *dns_rr_txt_parse(void *rr, size_t rrsiz, unsigned short rdp, unsigned short rdlen, const void *data_, size_t datalen, unsigned int flags, enum dns_errno *err) {
	struct dns_rr_txt *txt		= rr;
	size_t limit			= DNS_RR_TXT_LIMITOF(rrsiz);
	const unsigned char *data	= data_;
	size_t dlen;

	txt->tc.count	= 0;
	txt->tc.length	= 0;
	txt->count	= 0;
	txt->length	= 0;

	while (rdp < datalen) {
		dlen	= 0xff & data[rdp++];

		if (~rdp < dlen || rdp + dlen > datalen)
			goto invalid;

		if (txt->count < limit) {
			memcpy(txt->data[txt->count].string, &data[rdp], dlen);
			txt->data[txt->count].string[dlen]	= '\0';
			txt->data[txt->count].length		= dlen;

			txt->count++;
			txt->length	+= dlen;
		} else {
			txt->tc.count++;
			txt->tc.length	+= dlen;
		}

		rdp	+= dlen;
	}

	return txt;
invalid:
	*err	= DNS_EINVALID;

	return 0;
} /* dns_rr_txt_parse() */


size_t dns_rr_txt_comp(void *dst_, size_t dstsiz, const void *rr, struct dns_dict *dict, const void *data, size_t datalen, enum dns_errno *err) {
	unsigned char *dst		= dst_;
	size_t dstp			= 0;
	const struct dns_rr_txt *txt	= rr;
	unsigned i;

	for (i = 0; i < txt->count; i++) {
		if (dstp < dstsiz) {
			dst[dstp]	= txt->data[i].length;
			memcpy(&dst[dstp + 1], txt->data[i].string, MIN(dstsiz - (dstp + 1), txt->data[i].length));
		}

		dstp	+= 1 + txt->data[i].length;
	}

	return dstp;
} /* dns_rr_txt_comp() */


size_t dns_rr_txt_concat(void *dst_, size_t dstsiz, struct dns_rr_txt *txt, unsigned char a, unsigned char b) {
	unsigned char *dst	= dst_;
	size_t dstp		= 0;
	unsigned i;

	for (i = 0; a + i < txt->count && i < b; i++) {
		if (dstp < dstsiz)
			memcpy(&dst[dstp], txt->data[a + i].string, MIN(dstsiz - dstp, txt->data[a + i].length));

		dstp	+= txt->data[a + i].length;
	}

	if (dstsiz > 0)
		dst[DNS__MIN(dstp, dstsiz - 1)]	= '\0';

	return dstp;
} /* dns_rr_txt_concat() */


struct dns_rr_txt *dns_rr_txt_printf(void *rr, size_t rrsiz, const char *fmt, ...) {
	struct dns_rr_txt *txt	= rr;
	struct dns_rr_txt_data *data;
	unsigned char *dst	= (void *)&txt->data[0];
	unsigned char *dstp;
	size_t limit		= DNS_RR_TXT_LIMITOF(rrsiz) * (sizeof txt->data[0].string - 1);
	va_list ap;
	size_t len;
	unsigned i;

	va_start(ap, fmt);

	len	= vsnprintf((char *)dst, limit, fmt, ap);
	
	va_end(ap);

	if (-1 == len)
		return 0;

	txt->count	= howmany(MIN(len, limit), sizeof txt->data[0].string - 1);
	txt->length	= MIN(len, limit);

	txt->tc.count	= howmany(len, sizeof txt->data[0].string - 1) - txt->count;
	txt->tc.length	= len - txt->length;

	for (i = 0, dstp = dst; i < txt->count; i++) {
		data	= &txt->data[i];

		dstp		= memmove(&data->string[0], dstp, &dst[txt->length] - dstp);
		data->length	= MIN(&dst[txt->length] - dstp, DNS_RR_TXT_DATA_MAXSIZE);
		dstp		+= data->length;
	}

	for (i = 0; i < txt->count; i++)
		txt->data[i].string[txt->data[i].length]	= '\0';

	return txt;
} /* dns_rr_txt_printf() */


